/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     traverse_nr_common.c                                           */
/*                                                                          */
/* description:                                                             */
/*           nonrecursive mesh traversal, 2d/3d common routines             */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Univesitaet Bremen                                           */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

void fill_elinfo(int ichild, const EL_INFO *elinfo_old, EL_INFO *elinfo);

/*--------------------------------------------------------------------------*/
/*   List of static variables for stack-based traversal                     */
/*--------------------------------------------------------------------------*/
/*   traverse_mesh:       save pointer to current mesh		            */
/*   traverse_level:      save level from traverse_first() call		    */
/*   traverse_fill_flag:  save fill flag for traversal   		    */
/*   traverse_mel:        current macro element during traversal	    */
/*   stack_size:          current stack size                     	    */
/*   stack_used:          current level of hierarchy recursion   	    */
/*   elinfo_stack:        el_info hierarchy                      	    */
/*   info_stack:          describe path through hierarchy tree              */
/*                        =0: leaf of current hierarchy          	    */
/*                        =1: next element is child[0] of this one 	    */
/*                        =2: next element is child[1] of this one 	    */
/*   save_elinfo_stack:   backup for elinfo_stack during traverse_neighbour */
/*   save_info_stack:     backup for info_stack during traverse_neighbour   */
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#if 0
TRAVERSE_STACK *get_traverse_stack(void)
{
  FUNCNAME("get_traverse_stack");

  TRAVERSE_STACK *stack = MEM_ALLOC(1, TRAVERSE_STACK);

  stack->traverse_mesh = nil;
  stack->stack_size = 0;
  stack->stack_used = 0;
  stack->elinfo_stack = nil;
  stack->info_stack = nil;
  stack->save_elinfo_stack = nil;
  stack->save_info_stack = nil;
  stack->save_stack_used = 0;
  stack->el_count = 0;

  return(stack);
}


void free_traverse_stack(TRAVERSE_STACK *stack)
{
  FUNCNAME("free_traverse_stack");

  if (!stack) {
    ERROR("stack==nil ???\n");
    return;
  }

  if (stack->stack_size > 0) {
    MEM_FREE(stack->elinfo_stack, stack->stack_size, EL_INFO);
    MEM_FREE(stack->info_stack, stack->stack_size, U_CHAR);
    MEM_FREE(stack->save_elinfo_stack, stack->stack_size, EL_INFO);
    MEM_FREE(stack->save_info_stack, stack->stack_size, U_CHAR);
  }
  MEM_FREE(stack, 1, TRAVERSE_STACK);

  return;
}
#else
static TRAVERSE_STACK *free_stack = nil;

TRAVERSE_STACK *get_traverse_stack(void)
{
  FUNCNAME("get_traverse_stack");
  TRAVERSE_STACK *stack;

  if (!free_stack)
  {
    stack = MEM_ALLOC(1, TRAVERSE_STACK);

    stack->traverse_mesh = nil;
    stack->stack_size = 0;
    stack->stack_used = 0;
    stack->elinfo_stack = nil;
    stack->info_stack = nil;
    stack->save_elinfo_stack = nil;
    stack->save_info_stack = nil;
    stack->save_stack_used = 0;
    stack->el_count = 0;
    stack->next = nil;
  }
  else
  {
    stack = free_stack;
    free_stack = free_stack->next;
    stack->next = nil;
  }
  return(stack);
}


void free_traverse_stack(TRAVERSE_STACK *stack)
{
  FUNCNAME("free_traverse_stack");

  if (!stack) {
    ERROR("stack==nil ???\n");
    return;
  }

  stack->next = free_stack;
  free_stack = stack;

  return;
}
#endif
/*--------------------------------------------------------------------------*/

static void enlarge_traverse_stack(TRAVERSE_STACK *stack)
{
  FUNCNAME("enlarge_traverse_stack");
  int     i;
  int     new_stack_size = stack->stack_size + 10;

  stack->elinfo_stack = MEM_REALLOC(stack->elinfo_stack, stack->stack_size,
				    new_stack_size, EL_INFO);
  if (stack->stack_size > 0)
    for (i=stack->stack_size; i<new_stack_size; i++) 
      stack->elinfo_stack[i].fill_flag = stack->elinfo_stack[0].fill_flag;

  stack->info_stack = MEM_REALLOC(stack->info_stack, stack->stack_size,
				  new_stack_size, U_CHAR);
  stack->save_elinfo_stack = MEM_REALLOC(stack->save_elinfo_stack,
					 stack->stack_size, 
					 new_stack_size, EL_INFO);
  stack->save_info_stack   = MEM_REALLOC(stack->save_info_stack,
					 stack->stack_size,
					 new_stack_size, U_CHAR);

#if 0
  MSG("increase stack at %8X from %d to %d\n", stack,
      stack->stack_size, new_stack_size); 
  MSG("elinfo_stack      = %8X\n",stack->elinfo_stack);
  MSG("info_stack        = %8X\n",stack->info_stack);
  MSG("save_elinfo_stack = %8X\n",stack->save_elinfo_stack);
  MSG("save_info_stack   = %8X\n",stack->save_info_stack);
#endif

  stack->stack_size = new_stack_size;    
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*   return next element in tree traversal, starting from current one       */
/*                                                                          */
/*--------------------------------------------------------------------------*/

static EL_INFO *traverse_every_el_preorder(TRAVERSE_STACK *stack)
{
  FUNCNAME("traverse_every_el_preorder");
  EL *el;
  int i;
  
  if (stack->stack_used == 0)   /* first call */
  {
    stack->traverse_mel = stack->traverse_mesh->first_macro_el;
    if (stack->traverse_mel == nil)  return(nil);

    stack->stack_used = 1;
    fill_macro_info(stack->traverse_mesh, stack->traverse_mel,
		    stack->elinfo_stack+stack->stack_used);
    stack->info_stack[stack->stack_used] = 0;

    return(stack->elinfo_stack+stack->stack_used);
  }
  
  el = stack->elinfo_stack[stack->stack_used].el;

  /* go up in tree until we can go down again */
  while((stack->stack_used > 0) && 
	((stack->info_stack[stack->stack_used] >= 2) || (el->child[0]==nil)))
  {
    stack->stack_used--;
    el = stack->elinfo_stack[stack->stack_used].el;
  }


  /* goto next macro element */
  if (stack->stack_used < 1) {
    stack->traverse_mel = stack->traverse_mel->next;
    if (stack->traverse_mel == nil)  return(nil);

    stack->stack_used = 1;
    fill_macro_info(stack->traverse_mesh, stack->traverse_mel,
		    stack->elinfo_stack+stack->stack_used);
    stack->info_stack[stack->stack_used] = 0;

    return(stack->elinfo_stack+stack->stack_used);
  }


  /* go down tree */

  if (stack->stack_used >= stack->stack_size-1) enlarge_traverse_stack(stack);
  i = stack->info_stack[stack->stack_used];
  stack->info_stack[stack->stack_used]++;
  fill_elinfo(i, stack->elinfo_stack+stack->stack_used,
	      stack->elinfo_stack+stack->stack_used+1);
  stack->stack_used++;

  TEST_EXIT(stack->stack_used < stack->stack_size)
    ("stack_size=%d too small, level=%d\n",
     stack->stack_size, stack->elinfo_stack[stack->stack_used].level);

  stack->info_stack[stack->stack_used] = 0;
  
  return(stack->elinfo_stack+stack->stack_used); 
}


static EL_INFO *traverse_every_el_inorder(TRAVERSE_STACK *stack)
{
  EL *el;
  int i;

  if (stack->stack_used == 0)   /* first call */
  {
    stack->traverse_mel = stack->traverse_mesh->first_macro_el;
    if (stack->traverse_mel == nil)  return(nil);

    stack->stack_used = 1;
    fill_macro_info(stack->traverse_mesh, stack->traverse_mel,
		    stack->elinfo_stack+stack->stack_used);
    stack->info_stack[stack->stack_used] = 0;

    if (!stack->traverse_mel->el->child[0]) {
	return(stack->elinfo_stack+stack->stack_used);
    }
  }
  
  el = stack->elinfo_stack[stack->stack_used].el;

  /* go up in tree until we can go down again */    /* inorder !!! */
  i = 0;
  while((stack->stack_used > 0) && 
	((stack->info_stack[stack->stack_used] >= 2) || (el->child[0]==nil)))
  {
    stack->stack_used--;
    el = stack->elinfo_stack[stack->stack_used].el;
    ++i;
  }
  if (i && stack->stack_used >= 1) {
      return(stack->elinfo_stack+stack->stack_used);
  }


  /* goto next macro element */
  if (stack->stack_used < 1) {
    stack->traverse_mel = stack->traverse_mel->next;
    if (stack->traverse_mel == nil)  return(nil);

    stack->stack_used = 1;
    fill_macro_info(stack->traverse_mesh, stack->traverse_mel,
		    stack->elinfo_stack+stack->stack_used);
    stack->info_stack[stack->stack_used] = 0;

    if (!stack->traverse_mel->el->child[0]){
	return(stack->elinfo_stack+stack->stack_used);
    }
  }


  /* go down tree */

  while((stack->elinfo_stack+stack->stack_used)->el->child[0]
	&& (stack->info_stack[stack->stack_used] < 2)) {
    if (stack->stack_used >= stack->stack_size-1)
      enlarge_traverse_stack(stack);
    i = stack->info_stack[stack->stack_used];
    stack->info_stack[stack->stack_used]++;
    fill_elinfo(i, stack->elinfo_stack+stack->stack_used,
		stack->elinfo_stack+stack->stack_used+1);
    stack->stack_used++;
    stack->info_stack[stack->stack_used] = 0;
  }

  return(stack->elinfo_stack+stack->stack_used); 
}


static EL_INFO *traverse_every_el_postorder(TRAVERSE_STACK *stack)
{
  FUNCNAME("traverse_every_el_postorder");
  EL *el;
  int i;

  INFO(0,2)("\n");

  if (stack->stack_used == 0)   /* first call */
  {
    stack->traverse_mel = stack->traverse_mesh->first_macro_el;
    if (stack->traverse_mel == nil)  return(nil);

    stack->stack_used = 1;
    fill_macro_info(stack->traverse_mesh, stack->traverse_mel,
		    stack->elinfo_stack+stack->stack_used);
    stack->info_stack[stack->stack_used] = 0;

    if (!stack->traverse_mel->el->child[0])
	return(stack->elinfo_stack+stack->stack_used);
  }
  
  el = stack->elinfo_stack[stack->stack_used].el;

  /* go up in tree until we can go down again */          /* postorder!!! */
  while((stack->stack_used > 0) && 
	((stack->info_stack[stack->stack_used] >= 3) || (el->child[0]==nil)))
  {
    stack->stack_used--;
    el = stack->elinfo_stack[stack->stack_used].el;
  }


  /* goto next macro element */
  if (stack->stack_used < 1) {
    stack->traverse_mel = stack->traverse_mel->next;
    if (stack->traverse_mel == nil)  return(nil);

    stack->stack_used = 1;
    fill_macro_info(stack->traverse_mesh, stack->traverse_mel,
		    stack->elinfo_stack+stack->stack_used);
    stack->info_stack[stack->stack_used] = 0;

/*    return(stack->elinfo_stack+stack->stack_used); */
  }


  /* go down tree */

  while((stack->elinfo_stack+stack->stack_used)->el->child[0]
	&& (stack->info_stack[stack->stack_used] < 2)) {
    if (stack->stack_used >= stack->stack_size-1)
      enlarge_traverse_stack(stack);
    i = stack->info_stack[stack->stack_used];
    stack->info_stack[stack->stack_used]++;
    fill_elinfo(i, stack->elinfo_stack+stack->stack_used,
		stack->elinfo_stack+stack->stack_used+1);
    stack->stack_used++;
    stack->info_stack[stack->stack_used] = 0;
  }
  
  stack->info_stack[stack->stack_used]++;      /* postorder!!! */
  return(stack->elinfo_stack+stack->stack_used); 
}


static EL_INFO *traverse_leaf_el(TRAVERSE_STACK *stack)
{
  FUNCNAME("traverse_leaf_el");
  EL *el;
  int i;
  
  if (stack->stack_used == 0)   /* first call */
  {
    stack->traverse_mel = stack->traverse_mesh->first_macro_el;
    if (stack->traverse_mel == nil)  return(nil);

    stack->stack_used = 1;
    fill_macro_info(stack->traverse_mesh, stack->traverse_mel,
		    stack->elinfo_stack+stack->stack_used);
    stack->info_stack[stack->stack_used] = 0;

    el = stack->elinfo_stack[stack->stack_used].el;
    if ((el == nil) || (el->child[0] == nil)) {
      return(stack->elinfo_stack+stack->stack_used);
    }
  }
  else
  {
    el = stack->elinfo_stack[stack->stack_used].el;

    /* go up in tree until we can go down again */
    while((stack->stack_used > 0) && 
	  ((stack->info_stack[stack->stack_used] >= 2) || (el->child[0]==nil)))
    {
      stack->stack_used--;
      el = stack->elinfo_stack[stack->stack_used].el;
    }

    /* goto next macro element */
    if (stack->stack_used < 1) {
      stack->traverse_mel = stack->traverse_mel->next;
      if (stack->traverse_mel == nil)  return(nil);

      stack->stack_used = 1;
      fill_macro_info(stack->traverse_mesh, stack->traverse_mel,
		      stack->elinfo_stack+stack->stack_used);
      stack->info_stack[stack->stack_used] = 0;

      el = stack->elinfo_stack[stack->stack_used].el;
      if ((el == nil) || (el->child[0] == nil)) {
	return(stack->elinfo_stack+stack->stack_used);
      }
    }
  }

  /* go down tree until leaf */
  while (el->child[0])
  {
    if (stack->stack_used >= stack->stack_size-1)
      enlarge_traverse_stack(stack);
    i = stack->info_stack[stack->stack_used];
    el = el->child[i];
    stack->info_stack[stack->stack_used]++;
    fill_elinfo(i, stack->elinfo_stack+stack->stack_used,
		stack->elinfo_stack+stack->stack_used+1);
    stack->stack_used++;
    

    TEST_EXIT(stack->stack_used < stack->stack_size)
      ("stack_size=%d too small, level=(%d,%d)\n",
       stack->stack_size, stack->elinfo_stack[stack->stack_used].level);

    stack->info_stack[stack->stack_used] = 0;
  }

  return(stack->elinfo_stack+stack->stack_used); 

}

static EL_INFO *traverse_leaf_el_level(TRAVERSE_STACK *stack)
{
  FUNCNAME("traverse_leaf_el_level");

  ERROR_EXIT("not yet");
  return(nil);
}

static EL_INFO *traverse_el_level(TRAVERSE_STACK *stack)
{
  FUNCNAME("traverse_el_level");
  ERROR_EXIT("not yet");
  return(nil);
}

static EL_INFO *traverse_mg_level(TRAVERSE_STACK *stack)
{
  FUNCNAME("traverse_mg_level");
  ERROR_EXIT("not yet");
  return(nil);
}

/*--------------------------------------------------------------------------*/

#define CALL_EVERY_EL \
  (CALL_EVERY_EL_PREORDER | CALL_EVERY_EL_INORDER | CALL_EVERY_EL_POSTORDER)


const EL_INFO *traverse_next(TRAVERSE_STACK *stack, const EL_INFO *elinfo_old)
{
  FUNCNAME("traverse_next");
  EL_INFO       *elinfo=nil;

  if (stack->stack_used) {
    TEST_EXIT(elinfo_old == stack->elinfo_stack+stack->stack_used)
      ("invalid old elinfo\n");
  }
  else {
    TEST_EXIT(elinfo_old == nil)("invalid old elinfo != nil\n");
  }

  if (stack->traverse_fill_flag & CALL_LEAF_EL)
    elinfo = traverse_leaf_el(stack);
  else
    if (stack->traverse_fill_flag & CALL_LEAF_EL_LEVEL)
      elinfo = traverse_leaf_el_level(stack);
    else if (stack->traverse_fill_flag & CALL_EL_LEVEL)
      elinfo = traverse_el_level(stack);
    else if (stack->traverse_fill_flag & CALL_MG_LEVEL)
      elinfo = traverse_mg_level(stack);
    else
      if (stack->traverse_fill_flag & CALL_EVERY_EL_PREORDER)
	elinfo = traverse_every_el_preorder(stack);
      else if (stack->traverse_fill_flag & CALL_EVERY_EL_INORDER)
	elinfo = traverse_every_el_inorder(stack);
      else if (stack->traverse_fill_flag & CALL_EVERY_EL_POSTORDER)
	elinfo = traverse_every_el_postorder(stack);
      else
	ERROR_EXIT("invalid traverse_flag: %8x\n", stack->traverse_fill_flag);

  if (elinfo)
    stack->el_count++;
  else {
    /* MSG("total element count:%d\n",stack->el_count); */
  }

  return(elinfo);
}

#undef CALL_EVERY_EL



/*--------------------------------------------------------------------------*/
/*   traverse_first:                                 		            */
/*   ---------------                                    		    */
/*   initiate traversal and return first el_info                            */
/*--------------------------------------------------------------------------*/

const EL_INFO *traverse_first(TRAVERSE_STACK *stack, 
			      MESH *mesh, int level, FLAGS fill_flag)
{
  FUNCNAME("traverse_first");
  int i;

  if (!stack)
  {
    ERROR("no traverse stack\n");
    return(nil);
  }

  stack->traverse_mesh      = mesh;
  stack->traverse_level     = level; 
  stack->traverse_fill_flag = fill_flag; 

  if (stack->stack_size < 1)
    enlarge_traverse_stack(stack);

  for (i=0; i<stack->stack_size; i++)
    stack->elinfo_stack[i].fill_flag = fill_flag & FILL_ANY;

  stack->elinfo_stack[0].mesh = stack->elinfo_stack[1].mesh = mesh;

  if (fill_flag & CALL_LEAF_EL_LEVEL) {
    TEST_EXIT(level >= 0)("invalid level: %d\n",level);
  }

  stack->traverse_mel = nil;
  stack->stack_used   = 0;
  stack->el_count     = 0;

  return(traverse_next(stack, nil));
}


/*--------------------------------------------------------------------------*/
/*   test_traverse_fct:                              		            */
/*   ------------------                                 		    */
/*   display information about an element from EL and EL_INFO structures    */
/*--------------------------------------------------------------------------*/

static void test_traverse_nr_fct(const EL_INFO *elinfo)
{
  FUNCNAME("test_traverse2_fct");
  EL      *el = elinfo->el;

  MSG("\n");
  MSG("traversing element %d: at %8X -----------------------\n",
      INDEX(el), el);

  print_msg("level:        %3d\n",elinfo->level);

}

/*--------------------------------------------------------------------------*/
/*   test_traverse_nr:                                		            */
/*   -----------------                                   		    */
/*   display information about all elements from EL and EL_INFO structures  */
/*   by calling traverse with test_traverse_fct().      		    */
/*--------------------------------------------------------------------------*/

void test_traverse_nr(MESH *mesh, int level, FLAGS fill_flag)
{
  FUNCNAME("test_traverse_nr");
  const EL_INFO *elinfo;
  TRAVERSE_STACK *stack;

  MSG("with level    : %3d\n", level);
  MSG("with fill_flag:");
  if (fill_flag & FILL_ANY) {
    if (fill_flag & FILL_COORDS)      print_msg(" FILL_COORDS");
    if (fill_flag & FILL_NEIGH)       print_msg(" FILL_NEIGH");
    if (fill_flag & FILL_BOUND)       print_msg(" FILL_BOUND");
    if (fill_flag & FILL_OPP_COORDS)  print_msg(" FILL_OPP_COORDS");
  }
  else print_msg(" none");
  print_msg("\n");


  stack = get_traverse_stack();

  elinfo = traverse_first(stack, mesh, level, fill_flag);
  while (elinfo)
  {
    test_traverse_nr_fct(elinfo);

    elinfo = traverse_next(stack, elinfo);
  }

  free_traverse_stack(stack);

  MSG("done.\n");
 }
